from dataclasses import dataclass
from unittest.mock import patch

import pytest
from dyacon import Load
from dyacon.loaders.external import GoogleCloudSecretLoader
from dyacon.loaders.external.google import SecretManagerServiceClient
from dyacon.read import dataclass_from_dict


def test_env_loader():
    @dataclass
    class Config:
        key1: str = Load()

    config_dict = {'key1': '${ENV1}'}

    with patch('os.getenv', return_value='value1'):
        assert dataclass_from_dict(Config, config_dict).key1 == 'value1'


def test_env_loader_raise():
    @dataclass
    class Config:
        key1: str = Load()

    config_dict = {'key1': '${ENV1}'}

    with patch('os.getenv', return_value=None):
        with pytest.raises(
            ValueError,
            match='Missing required environment variable ENV1',
        ):
            dataclass_from_dict(Config, config_dict)


def test_env_loader_default_value():
    @dataclass
    class Config:
        key1: str = Load()

    config_dict = {'key1': '${ENV1:default}'}

    with patch('os.getenv', return_value=None):
        assert dataclass_from_dict(Config, config_dict).key1 == 'default'


def test_env_loader_several_groups():
    @dataclass
    class Config:
        key1: str = Load()

    config_dict = {'key1': 'hello-${ENV1}${ENV2}'}

    with patch.dict('os.environ', {'ENV1': 'value1', 'ENV2': 'value2'}):
        assert dataclass_from_dict(Config, config_dict).key1 == (
            'hello-value1value2'
        )


@pytest.mark.skipif(
    SecretManagerServiceClient is not None,
    reason=(
        'google-cloud-secret-manager is installed, '
        'but secret_name and project_id are not real'
    ),
)
def test_google_loader_default():
    @dataclass
    class Config:
        key1: str = Load([GoogleCloudSecretLoader()])

    config_dict = {'key1': '!{secret_name:project_id:default}'}

    assert dataclass_from_dict(Config, config_dict).key1 == 'default'


@pytest.mark.skipif(
    SecretManagerServiceClient is not None,
    reason=(
        'google-cloud-secret-manager is installed, '
        'but secret_name and project_id are not real'
    ),
)
def test_google_loader_without_default():
    @dataclass
    class Config:
        key1: str = Load([GoogleCloudSecretLoader()])

    config_dict = {'key1': '!{secret_name:project_id}'}

    with pytest.raises(
        ValueError,
        match='Missing required google cloud secret manager',
    ):
        dataclass_from_dict(Config, config_dict)


def test_not_required():
    @dataclass
    class Config:
        key1: str = Load()

    config_dict = {}

    assert dataclass_from_dict(Config, config_dict).key1 is None


def test_required():
    @dataclass
    class Config:
        key1: str = Load(required=True)

    config_dict = {}

    with pytest.raises(ValueError, match='key1 field cannot be empty'):
        dataclass_from_dict(Config, config_dict)


def test_default():
    @dataclass
    class Config:
        key1: str = Load(default='default')

    config_dict = {}

    assert dataclass_from_dict(Config, config_dict).key1 == 'default'


def test_default_factory():
    @dataclass
    class Config:
        key1: str = Load(default_factory=lambda: 'default')

    config_dict = {}

    assert dataclass_from_dict(Config, config_dict).key1 == 'default'



import os
import re
from abc import ABC, abstractmethod
from typing import List, Optional, Pattern, Callable, Any


class Loader(ABC):
    @property
    @abstractmethod
    def pattern(self) -> Pattern[str]:
        raise NotImplementedError

    @abstractmethod
    def load(self, config_content: str) -> str:
        raise NotImplementedError


# fmt: off
ENV_LOADER_PATTERN = re.compile(
    r'\${'
    r'(?P<name>[^}^{:]+)'
    r'(?P<separator>:?)'
    r'(?P<default>.*?)'
    r'}'
)
# fmt: on


class EnvLoader(Loader):
    def __init__(self) -> None:
        self._pattern = ENV_LOADER_PATTERN

    @property
    def pattern(self) -> Pattern[str]:
        return self._pattern

    def load(self, config_content: str) -> str:
        findings = list(self.pattern.finditer(config_content))

        for match in reversed(findings):
            env_name = match.group('name')
            has_default = match.group('separator') == ':'

            value = os.getenv(env_name)

            if value is None:
                if not has_default:
                    raise ValueError(
                        f'Missing required environment variable {env_name}'
                    )
                value = match.group('default')
            span_min, span_max = match.span()
            config_content = (
                f'{config_content[:span_min]}'
                f'{value}'
                f'{config_content[span_max:]}'
            )
        return config_content


class Load:
    """
    for extending just inherit from this class and override __init__ method
    or use lambda function
    examples:
    >>> MyLoad1 = lambda: Load(loaders=[MyLoader()])
    >>> class MyLoad2(Load):
    ...     def __init__(
    ...          self,
    ...          loaders: Optional[List[Loader]] = None,
    ...          *,
    ...          default: Optional[str] = None,
    ...          default_factory: Optional[Callable[[], Any]] = None,
    ...          required: bool = False,
    ...      ) -> None:
    ...          super().__init__(
    ...              loaders,
    ...              default=default,
    ...              default_factory=default_factory,
    ...              required=required,
    ...          )
    ...          self.loaders.append(MyLoader())  # <-- here is your new loader
    """

    def __init__(
        self,
        loaders: Optional[List[Loader]] = None,
        *,
        default: Optional[str] = None,
        default_factory: Optional[Callable[[], Any]] = None,
        required: bool = False,
    ) -> None:
        self.loaders = [EnvLoader()]
        self.required = required
        self.default = default
        self.default_factory = default_factory
        if loaders:
            self.loaders.extend(loaders)

    def load(self, config_content: str) -> str:
        for loader in self.loaders:
            if list(loader.pattern.finditer(config_content)):
                return loader.load(config_content)
        return config_content

    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        return instance.__dict__[self.name]

    def __set__(self, instance, value):
        if self is not value:
            instance.__dict__[self.name] = self.load(value)
        else:
            if self.required:
                raise ValueError(f'{self.name} field cannot be empty')
            if self.default_factory:
                instance.__dict__[self.name] = self.default_factory()
            else:
                instance.__dict__[self.name] = self.default
